import re
from datetime import datetime
from pathlib import Path
from typing import Iterable, List, Tuple

from dateutil.parser import parse as parse_date

from neoteroi.contribs.domain import ContributionsReader, Contributor
from neoteroi.markdown.commands import Command


class GitContributionsReader(ContributionsReader):

    _name_email_rx = re.compile(r"(?P<name>[\w\s]+)\s<(?P<email>[^\>]+)>")

    def _parse_name_and_email(self, name_and_email) -> Tuple[str, str]:
        match = self._name_email_rx.search(name_and_email)
        if match:
            name = match.groupdict()["name"]
            email = match.groupdict()["email"]
        else:
            name, email = ""
        return name, email

    def parse_committers(self, output: str) -> Iterable[Contributor]:
        for line in output.splitlines():
            count, name_and_email = line.split("\t")
            name, email = self._parse_name_and_email(name_and_email)
            yield Contributor(name, email, int(count))

    def get_contributors(self, file_path: Path) -> List[Contributor]:
        """
        Obtains the list of contributors for a file with the given path,
        using the Git CLI.
        """
        command = Command(f'git shortlog --summary --numbered --email "{file_path}"')

        result = command.execute()
        return list(self.parse_committers(result))

    def get_last_commit_date(self, file_path: Path) -> datetime:
        """Reads the last commit on a file."""
        command = Command(f'git log -1 --pretty="format:%ci" "{file_path}"')

        result = command.execute()
        return parse_date(result)



"""
This module provide a plugin that can be used to generate last modified time and
contributors list for each page, assuming that:

- the MkDocs site is built in a Git repository.
- the Git CLI can be used during the build.

"""
import logging
from datetime import datetime
from pathlib import Path
from typing import List

from mkdocs.config import config_options as c
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import File
from mkdocs.structure.pages import Page

from neoteroi.contribs.domain import ContributionsReader, Contributor
from neoteroi.contribs.git import GitContributionsReader
from neoteroi.contribs.html import ContribsViewOptions, render_contribution_stats

logger = logging.getLogger("MARKDOWN")


class ContribsPlugin(BasePlugin):
    _contribs_reader: ContributionsReader
    config_scheme = (
        ("contributors_label", c.Type(str, default="Contributors")),
        ("last_modified_label", c.Type(str, default="Last modified on")),
        ("last_modified_time", c.Type(bool, default=True)),
        ("time_format", c.Type(str, default="%Y-%m-%d %H:%M:%S")),
        ("contributors", c.Type(list, default=[])),
    )

    def __init__(self) -> None:
        super().__init__()
        self._contribs_reader = GitContributionsReader()

    def _get_contributors(self, page_file: File) -> List[Contributor]:
        contributors = self._contribs_reader.get_contributors(
            Path("docs") / page_file.src_path
        )
        info = self.config.get("contributors")

        if not info:
            return contributors

        for contributor in contributors:
            contributor_info = next(
                (item for item in info if item.get("email") == contributor.email), None
            )
            if contributor_info:
                contributor.image = contributor_info.get("image")

        return contributors

    def _get_last_commit_date(self, page_file: File) -> datetime:
        return self._contribs_reader.get_last_commit_date(
            Path("docs") / page_file.src_path
        )

    def _set_contributors(self, markdown: str, page: Page) -> str:
        page_file = page.file
        last_commit_date = self._get_last_commit_date(page_file)
        contributors = self._get_contributors(page_file)
        return markdown + render_contribution_stats(
            contributors,
            last_commit_date,
            ContribsViewOptions(
                self.config["contributors_label"],
                self.config["last_modified_label"],
                self.config["last_modified_time"],
                self.config["time_format"],
            ),
        )

    def on_page_markdown(self, markdown, *args, **kwargs):
        try:
            markdown = self._set_contributors(markdown, kwargs["page"])
        except ValueError:
            logger.error(
                "Failed to display contributors list for page: %s",
                kwargs["page"].title,
            )
            pass
        return markdown



import logging
import subprocess

logger = logging.getLogger("MARKDOWN")


class Command:
    def __init__(self, command: str):
        self._command = None
        self.command = command

    @property
    def command(self):
        if not self._command:
            raise TypeError("Missing command")
        return self._command

    @command.setter
    def command(self, value):
        self._command = value

    def execute(self) -> str:
        logger.debug(f"\nNeoteroi Contrib; executing command:\n  {self.command}\n")
        p = subprocess.Popen(
            self.command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if p.stdout is None:
            raise TypeError("Missing subprocess stdout")

        output = p.stdout.read()
        error_output = p.stderr.read() if p.stderr else None

        try:
            output = output.decode("utf8")
        except UnicodeDecodeError:
            output = output.decode("ISO-8859-1")

        if error_output:
            raise RuntimeError(
                f"Process failed with return code: "
                f"{p.returncode}.\nOutput: {output}"
            )

        if p.returncode is not None and p.returncode != 0:
            raise RuntimeError(
                f"Process failed with return code: "
                f"{p.returncode}.\nOutput: {output}"
            )

        if output:
            logger.debug(f"Neoteroi Contrib; got output:\n  {output}\n")
        return output



from dataclasses import dataclass
from datetime import date, datetime
from typing import Iterable, List, Optional, Union

from dateutil.parser import parse as parse_date

from neoteroi.projects.timeutil import parse_lasts


@dataclass(frozen=True)
class Event:
    title: str
    description: Optional[str] = None
    time: Optional[datetime] = None
    icon: Optional[str] = None

    def __post_init__(self):
        # Note: datetimes are currently not supported, only the date component is kept
        if isinstance(self.time, datetime):
            object.__setattr__(self, "time", self.time.date())

    @classmethod
    def from_obj(cls, obj):
        return cls(
            obj.get("title") or "",
            obj.get("description") or "",
            _parse_optional_datetime(obj.get("time")),
            icon=obj.get("icon"),
        )


def _parse_optional_datetime(value: Union[None, datetime, str]) -> Optional[datetime]:
    if isinstance(value, date):
        # YAML parses dates automatically
        return value

    if value:
        return parse_date(value)
    return None


def _parse_optional_date(value: Union[None, date, str]) -> Optional[date]:
    if isinstance(value, date):
        # YAML parses dates automatically
        return value

    if value:
        return parse_date(value)
    return None


def _resolve_activities(
    activities, preceding_date: Optional[date] = None
) -> Iterable["Activity"]:
    """
    Iterates through a list of objects representing activities and yields
    activities, handling automatic start date when the preceding activity specifies one.
    """
    for item in activities:
        activity = Activity.from_obj(item, preceding_date)
        preceding_date = activity.end or activity.start
        yield activity


@dataclass(frozen=True)
class Activity:
    title: str
    start: Optional[date] = None
    end: Optional[date] = None
    description: Optional[str] = None
    activities: Optional[List["Activity"]] = None
    events: Optional[List[Event]] = None
    hidden: Optional[bool] = None

    def __post_init__(self):
        # Note: datetimes are currently not supported, only the date component is kept
        if isinstance(self.start, datetime):
            object.__setattr__(self, "start", self.start.date())
        if isinstance(self.end, datetime):
            object.__setattr__(self, "end", self.end.date())

    def iter_activities(self, include_self: bool = True) -> Iterable["Activity"]:
        """
        Yields self and all descendant activities.
        """
        if include_self:
            yield self

        if self.activities:
            for activity in self.activities:
                yield from activity.iter_activities()

    def iter_events(self, include_self: bool = True) -> Iterable[Event]:
        """
        Yields all events in the activity and descendants.
        """
        for activity in self.iter_activities(include_self):
            if activity.events:
                yield from activity.events

    def iter_dates(self) -> Iterable[date]:
        if self.events:
            for event in self.events:
                if event.time:
                    yield event.time
        if self.activities:
            for activity in self.activities:
                if activity.start:
                    yield activity.start
                if activity.end:
                    yield activity.end

    def _get_event_date(self, fn) -> Optional[date]:
        if self.events:
            dates = [event.time for event in self.events if event.time is not None]
            return fn(dates) if dates else None
        return None

    def get_first_event_date(self) -> Optional[date]:
        return self._get_event_date(min)

    def get_last_event_date(self) -> Optional[date]:
        return self._get_event_date(max)

    def get_overall_start(self) -> Optional[date]:
        """
        Returns the start date of this activity, including all sub and descendants
        activities.
        """
        all_starts = [
            activity.start
            for activity in self.iter_activities()
            if activity.start is not None
        ] + [event.time for event in self.iter_events() if event.time is not None]
        return min(all_starts)

    def get_overall_end(self) -> Optional[date]:
        """
        Returns the end date of this activity, including all sub and descendants
        activities.
        """
        all_ends = [
            activity.end
            for activity in self.iter_activities()
            if activity.end is not None
        ] + [event.time for event in self.iter_events() if event.time is not None]
        return max(all_ends)

    @classmethod
    def from_obj(cls, obj, preceding_date: Optional[date] = None):
        if isinstance(obj, str):
            return cls(title=obj)
        if not isinstance(obj, dict):
            raise TypeError("Expected a str or a dict.")

        title = obj.get("title", "")
        description = obj.get("description")
        start = _parse_optional_date(obj.get("start", preceding_date))
        end = _parse_optional_date(obj.get("end"))
        skip = obj.get("skip")
        if skip:
            lasts = skip
            hidden = True
        else:
            lasts = obj.get("lasts")
            hidden = obj.get("hidden")
        child_activities = obj.get("activities") or []
        events = obj.get("events")
        if lasts:
            if start is None:
                start = date.today() if preceding_date is None else preceding_date
            end = start + parse_lasts(lasts)
        if preceding_date is None:
            preceding_date = end or start

        if child_activities and end is None:
            # avoid displaying a parent activity that would last anyway from the
            # beginning to the end of the last descendant
            hidden = True

        return cls(
            title,
            start,
            end,
            description=description,
            activities=list(_resolve_activities(child_activities, end or start))
            if child_activities
            else None,
            events=[Event.from_obj(item) for item in events] if events else None,
            hidden=hidden,
        )


@dataclass(frozen=True)
class Plan(Activity):
    def iter_activities(self, include_self: bool = False) -> Iterable["Activity"]:
        yield from super().iter_activities(include_self)

    @classmethod
    def from_obj(cls, obj):
        if isinstance(obj, list):
            return cls(
                title="Plan",
                activities=[Activity.from_obj(item) for item in obj],
            )

        if not isinstance(obj, dict):
            raise TypeError("Expected a list or a dict.")

        plan_start = _parse_optional_date(obj.get("start"))
        activities = obj.get("activities") or []
        events = obj.get("events")

        return cls(
            title=obj.get("title") or "Plan",
            activities=[Activity.from_obj(item, plan_start) for item in activities]
            if activities
            else [],
            events=[Event.from_obj(item) for item in events] if events else [],
        )
